# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import inspect
from typing import Any, no_type_check


def is_forward_ref(annotation: Any) -> bool:
    """Returns True when the given annotation is a forward ref.
    """
    return issubclass(type(annotation), _forward_ref)


@no_type_check
def forward_ref(name: str) -> Any:
    """Generate a proxy type for a schema that is going to be defined
    at some point after the current statement.

    Examples:
      Schema "A" includes a reference to a yet-to-be-defined schema
      "B"::

        @schema
        class A:
            b: forward_ref("B")

        @schema
        class B:
            x: int

      ``forward_ref`` can be used as a type parameter to other types
      as well::

        @schema
        class A:
            b: List[forward_ref("B")]

        @schema
        class B:
            x: int

    Parameters:
      name: The name of the schema being referenced.  This name must
        eventually be bound in the scope of the call site.
    """
    try:
        caller_frame_info = inspect.stack()[1]
        return _forward_ref(f'_forward_ref("{name}")', (object,), {
            "_schema": None,
            "_name": name,
            "_namespace": caller_frame_info.frame.f_globals,
        })
    finally:
        del caller_frame_info


class _forward_ref(type):
    @no_type_check
    def lookup(cls):
        schema = cls._schema
        if schema is None:
            schema = cls._schema = cls._namespace[cls._name]
            del cls._name
            del cls._namespace

        return schema



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Here be Dragons.  OpenAPI and especially JSONSchema are not very
# good specs, but at least they're popular!

import ast
import inspect
from collections import defaultdict
from operator import itemgetter
from textwrap import dedent
from types import FunctionType, MethodType
from typing import (
    Any, Callable, Dict, List, Optional, Set, Tuple, Union, get_type_hints, no_type_check
)

from typing_inspect import get_origin, is_generic_type, is_typevar, is_union_type

from ..app import BaseApp
from ..http import UploadedFile
from ..router import get_route_parameters
from ..typing import Header, QueryParam, extract_optional_annotation, get_args
from ..validation import Field, dump_schema, field, is_schema, schema


@schema
class Contact:
    """Contact information for the exposed API.

    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#contact-object
    """

    name: Optional[str] = None
    url: Optional[str] = None
    email: Optional[str] = None


@schema
class License:
    """License information for the exposed API.

    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#license-object
    """

    name: str
    url: Optional[str] = None


@schema
class Metadata:
    """Metadata about the exposed API.

    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#infoObject
    """

    title: str
    description: str
    version: str

    terms_of_service: Optional[str] = field(response_name="termsOfService", default=None)
    contact: Optional[Contact] = None
    license: Optional[License] = None


@schema
class Schema:
    """Describes the type and attributes of a value.

    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#schemaObject
    """

    type: Optional[str] = None
    format: Optional[str] = None  # noqa
    description: Optional[str] = None
    required: Optional[List[str]] = None
    properties: Optional[Dict[str, Any]] = None
    nullable: Optional[bool] = None
    all_of: Optional[List[Dict[str, str]]] = field(response_name="allOf", default=None)
    any_of: Optional[List[Dict[str, str]]] = field(response_name="anyOf", default=None)
    items: Optional[Dict[str, str]] = None
    choices: Optional[List[str]] = field(response_name="enum", default=None)
    pattern: Optional[str] = None
    minimum: Optional[Union[int, float]] = None
    maximum: Optional[Union[int, float]] = None
    multiple_of: Optional[Union[int, float]] = field(response_name="multipleOf", default=None)
    min_length: Optional[int] = field(response_name="minLength", default=None)
    max_length: Optional[int] = field(response_name="maxLength", default=None)
    read_only: Optional[bool] = field(response_name="readOnly", default=None)
    write_only: Optional[bool] = field(response_name="writeOnly", default=None)
    ref: Optional[str] = field(response_name="$ref", default=None)


@schema
class Parameter:
    """Describes a single handler parameter.

    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#parameterObject
    """

    name: str
    in_: str = field(response_name="in", choices=["query", "header", "path", "cookie"])
    description: Optional[str] = None
    required: Optional[bool] = None
    deprecated: bool = False
    schema: Optional[Schema] = None


@schema
class APIKeySecurityScheme:
    """Describes an API key-based security scheme.

    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#securitySchemeObject
    """

    name: str = field(request_only=True)
    param_name: str = field(response_name="name")
    in_: str = field(response_name="in", choices=["query", "header", "cookie"])
    type: str = field(response_only=True, choices=["apiKey"], default="apiKey")


@schema
class HTTPSecurityScheme:
    """Describes an HTTP-based security scheme.

    https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.0.1.md#securitySchemeObject
    """

    name: str
    scheme: str = field(choices=["basic", "bearer"])
    type: str = field(response_only=True, choices=["http"], default="http")


#: The union of acceptable security schemes.
SecurityScheme = Union[
    APIKeySecurityScheme,
    HTTPSecurityScheme,
]

@no_type_check  # noqa
def generate_openapi_document(
        app: BaseApp,
        metadata: Metadata,
        security_schemes: List[SecurityScheme],
        default_security_scheme: Optional[str] = None,
) -> Dict[str, Any]:
    """Generate an OpenAPI v3 document from an application object and
    some metadata.
    """
    request_mime_types = {parser.mime_type for parser in app.parsers}
    response_mime_types = {renderer.mime_type for renderer in app.renderers}

    schemas: Dict[str, Schema] = {}
    paths: Dict[str, Dict[str, Any]] = defaultdict(dict)
    for method, routes in app.router._routes_by_method.items():
        method = method.lower()

        for route in routes:
            handler = route.handler
            if not isinstance(handler, (FunctionType, MethodType)):
                handler = handler.__call__  # type: ignore

            parameters = []
            request_schema_name = None
            annotations = get_type_hints(handler)
            route_template_parameters = get_route_parameters(route.template)
            for name, annotation in annotations.items():
                if name == "return":
                    continue

                is_optional, annotation = extract_optional_annotation(annotation)
                if name in route_template_parameters:
                    parameters.append(Parameter(
                        name, "path",
                        description=_get_annotation(handler, f"param_{name}_description"),
                        required=True,
                        schema=_generate_primitive_schema(annotation),
                    ))

                elif annotation is QueryParam:
                    parameters.append(Parameter(
                        name, "query",
                        description=_get_annotation(handler, f"param_{name}_description"),
                        required=not is_optional,
                        schema=Schema("string"),
                    ))

                elif annotation is Header:
                    parameters.append(Parameter(
                        name.replace("_", "-"), "header",
                        description=_get_annotation(handler, f"param_{name}_description"),
                        required=not is_optional,
                        schema=Schema("string"),
                    ))

                elif is_schema(annotation):
                    request_schema_name = _generate_schema(annotation, schemas)

            operation = paths[route.template][method] = {
                "tags": _get_annotation(handler, "tags", []),
                "operationId": route.name,
                "description": dedent(handler.__doc__ or "").rstrip(),
                "parameters": [dump_schema(param, sparse=True) for param in parameters],
                "deprecated": _get_annotation(handler, "deprecated", False),
                "responses": {"200": {"description": "A successful response.", "content": {}}},
            }

            if request_schema_name is not None:
                operation["requestBody"] = {"content": {}}
                for media_type in request_mime_types:
                    operation["requestBody"]["content"][media_type] = {
                        "schema": _make_schema_ref(request_schema_name),
                    }

                # Sort the media types for tools like the Swagger UI.
                operation["requestBody"]["content"] = _sort_dict(operation["requestBody"]["content"])

            # When the response annotation is a tuple of the form
            # [str, ...] then we assume that the responses will
            # contain custom status codes followed by the response
            # objects themselves.
            response_annotation = annotations.get("return")
            response_annotation_origin = get_origin(response_annotation)
            if response_annotation is not None and response_annotation_origin in _TUPLE_TYPES:
                arguments = get_args(response_annotation)
                if len(arguments) == 2 and arguments[0] is str and is_schema(arguments[1]):
                    response_annotation = arguments[1]

            if response_annotation is not None:
                if is_schema(response_annotation):
                    response_schema_name = _generate_schema(response_annotation, schemas)
                    for media_type in response_mime_types:
                        operation["responses"]["200"]["content"][media_type] = {
                            "schema": _make_schema_ref(response_schema_name),
                        }

                elif response_annotation_origin in _LIST_TYPES:
                    arguments = get_args(response_annotation)
                    if is_schema(arguments[0]):
                        response_schema_name = _generate_schema(arguments[0], schemas)
                        for media_type in response_mime_types:
                            operation["responses"]["200"]["content"][media_type] = {
                                "schema": {
                                    "type": "array",
                                    "items": _make_schema_ref(response_schema_name),
                                }
                            }

            status_codes = _extract_status_codes(handler)
            for status_code in status_codes:
                status_code_ob = operation["responses"][str(status_code)] = {}
                if status_code < 300 and status_code != 204:
                    status_code_ob.update(**operation["responses"]["200"])

                description = _get_annotation(handler, f"response_{status_code}_description")
                status_code_ob.update(description=description)

            # If the declared return type is a response-code-tuple and
            # the status code finder couldn't find a 200 status code,
            # then it should be safe to drop that code from the
            # responses object.
            if get_origin(annotations.get("return")) in _TUPLE_TYPES and 200 not in status_codes:
                del operation["responses"]["200"]

            # TODO: Add support for OAuth2 security scheme.
            # TODO: Make it possible to annotate that a handler
            # doesn't require a security scheme.
            if default_security_scheme is not None:
                operation["security"] = [{default_security_scheme: []}]

    return {
        "openapi": "3.0.1",
        "info": dump_schema(metadata, sparse=True),
        "paths": _sort_dict(paths),
        "components": {
            "schemas": _sort_dict({name: dump_schema(schema, sparse=True) for name, schema in schemas.items()}),
            "securitySchemes": _sort_dict({scheme.name: dump_schema(scheme, sparse=True) for scheme in security_schemes}),
        },
    }


@no_type_check
def _generate_schema(schema: Any, schemas: Dict[str, Schema]) -> str:
    name = f"{schema.__module__}.{schema.__name__}"
    if name in schemas:
        return name

    definition = Schema(
        "object",
        description=schema.__doc__,
        properties={},
        required=[],
    )

    for field in schema._FIELDS.values():  # noqa
        if field.request_name == field.response_name:
            field_names = [field.request_name]
        else:
            field_names = [field.request_name, field.response_name]

        for field_name in field_names:
            is_optional, field_schema = _generate_field_schema(field_name, field, schemas)
            if field_schema is not None:
                definition.properties[field_name] = field_schema

            if not is_optional:
                definition.required.append(field_name)

    schemas[name] = definition
    return name


@no_type_check
def _generate_field_schema(field_name: str, field: Field, schemas: Dict[str, Schema]) -> Tuple[bool, Schema]:
    is_optional, annotation = extract_optional_annotation(field.annotation)
    if is_schema(annotation):
        field_schema_name = _generate_schema(annotation, schemas)
        field_schema = Schema(ref=_make_ref_path(field_schema_name))

    elif is_generic_type(annotation):
        origin = get_origin(annotation)
        if origin in _LIST_TYPES:
            arguments = get_args(annotation)
            if arguments and is_schema(arguments[0]):
                item_schema_name = _generate_schema(arguments[0], schemas)
                field_schema = Schema("array", items=_make_schema_ref(item_schema_name))

            else:
                field_schema = _generate_primitive_schema(annotation)

        elif origin in _DICT_TYPES:
            # TODO: Add support for additionalFields.
            field_schema = _generate_primitive_schema(dict)

        else:  # pragma: no cover
            raise ValueError(f"Unsupported type {origin} for field {field.name!r}.")

    elif is_union_type(annotation):
        sub_schemas = []
        for arg in get_args(annotation):
            _, sub_schema = _generate_field_schema("", Field(annotation=arg), schemas)
            sub_schemas.append(dump_schema(sub_schema, sparse=True))

        field_schema = Schema(any_of=sub_schemas)

    else:
        field_schema = _generate_primitive_schema(annotation)

    if field_schema is not None:
        field_schema.description = field.description

        if field.request_name != field.response_name:
            if field_name == field.request_name:
                field_schema.write_only = True

            else:
                field_schema.read_only = True

        elif field.response_only:
            field_schema.read_only = True

        elif field.request_only:
            field_schema.write_only = True

        for option, value in field.validator_options.items():
            if option in Schema._FIELDS:
                setattr(field_schema, option, value)

    return is_optional, field_schema


@no_type_check
def _generate_primitive_schema(annotation: Any) -> Optional[Schema]:
    try:
        arguments = _PRIMITIVE_ANNOTATION_MAP[annotation]
        return Schema(*arguments)
    except KeyError:
        origin = get_origin(annotation)
        if origin in _LIST_TYPES:
            arguments = get_args(annotation)
            if not arguments or is_typevar(arguments[0]):
                return Schema("array", items=_ANY_VALUE)

            else:
                return Schema("array", items=_generate_primitive_schema(arguments[0]))

        # TODO: Add support for additionalFields.
        return Schema("string")


def _extract_status_codes(handler: Callable[..., Any]) -> List[int]:
    try:
        source = inspect.getsource(handler)
        finder = _StatusCodeFinder()
        return finder.find(ast.parse(dedent(source)))
    except OSError:  # pragma: no cover
        return []


def _make_ref_path(name: str) -> str:
    return f"#/components/schemas/{name}"


def _make_schema_ref(name: str) -> Dict[str, str]:
    return {"$ref": _make_ref_path(name)}


def _get_annotation(handler: Callable[..., Any], name: str, default: Any = None) -> Any:
    return getattr(handler, f"openapi_{name}", default)


def _sort_dict(data: Dict[Any, Any]) -> Dict[Any, Any]:
    # This relies on the ordered dict implementation in Py3.6+.
    return dict(sorted(data.items(), key=itemgetter(0)))


class _StatusCodeFinder(ast.NodeVisitor):
    """Finds usages of HTTP_* in an AST.
    """

    def __init__(self) -> None:
        self.status_codes: Set[int] = set()

    def find(self, tree: Any) -> List[int]:
        self.visit(tree)
        return sorted(list(self.status_codes))

    def visit_Name(self, node: Any) -> None:
        if node.id.startswith("HTTP_"):
            try:
                self.status_codes.add(int(node.id[len("HTTP_"):]))
            except ValueError:  # pragma: no cover
                pass


#: Maps primitive types to Schema parameters.
_PRIMITIVE_ANNOTATION_MAP = {
    UploadedFile: ["string", "binary"],
    int: ["integer", "int64"],
    str: ["string"],
    bool: ["boolean"],
    dict: ["object"],
    float: ["number", "double"],
    bytes: ["string", "binary"],
}

#: A schema that accepts any value whatsoever.  Used in generic types
#: w/o a type annotation (eg. List).
_ANY_VALUE = {
    "description": "Can be any value, including null.",
    "nullable": True,
}

_DICT_TYPES = {dict, Dict}
_LIST_TYPES = {list, List}
_TUPLE_TYPES = {tuple, Tuple}



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import os
import random
import string
from functools import partial
from io import BytesIO
from json import dumps as to_json
from typing import Any, BinaryIO, Callable, Dict, Optional, Tuple, Union, cast
from urllib.parse import urlencode

from ..app import BaseApp
from ..http import HTTP_200
from ..http.headers import Headers, HeadersDict
from ..http.query_params import ParamsDict, QueryParams
from ..http.request import Request
from ..http.response import Response
from ..typing import Environ
from .common import to_environ

try:
    from gunicorn.http.errors import NoMoreData  # type: ignore
except ImportError:  # pragma: no cover
    class NoMoreData(Exception):  # type: ignore
        pass


HTTP_METHODS = {"delete", "head", "get", "options", "patch", "post", "put"}


class TestResponse:
    """A wrapper around Response objects that adds a few additional
    helper methods for testing.

    Attributes:
      status: The response status line.
      status_code: The response status code as an integer.
      headers: The response headers.
      stream: The response data as a binary file.
      data: The response data as a string.
    """

    __slots__ = ["_response"]

    def __init__(self, response: Response) -> None:
        self._response = response

    @property
    def data(self) -> str:
        """Rewinds the output stream and returns all its data.
        """
        self._response.stream.seek(0)
        return self._response.stream.read().decode("utf-8")

    @property
    def status_code(self) -> int:
        """Returns the HTTP status code as an integer.
        """
        code, _, _ = self._response.status.partition(" ")
        return int(code)

    def json(self) -> Any:
        """Convert the response data to JSON.
        """
        return json.loads(self.data)

    def __getattr__(self, name: str) -> Any:
        return getattr(self._response, name)


FileSpec = Union[
    Tuple[str, BinaryIO],
    Tuple[str, str, BinaryIO],
]


class TestClient:
    """Test clients are used to simulate requests against an
    application instance.
    """

    __slots__ = ["app"]

    def __init__(self, app: BaseApp) -> None:
        self.app = app

    def request(
            self,
            method: str,
            path: str,
            headers: Optional[Union[HeadersDict, Headers]] = None,
            params: Optional[Union[ParamsDict, QueryParams]] = None,
            body: Optional[bytes] = None,
            data: Optional[Dict[str, str]] = None,
            files: Optional[Dict[str, FileSpec]] = None,
            json: Optional[Any] = None,
            auth: Optional[Callable[[Request], Request]] = None,
            prepare_environ: Optional[Callable[[Environ], Environ]] = None,
    ) -> TestResponse:
        """Simulate a request against the application.

        Raises:
          RuntimeError: If both 'data' and 'json' are provided.

        Parameters:
          method: The request method.
          path: The request path.
          headers: Optional request headers.
          params: Optional query params.
          body: An optional bytestring for the request body.
          data: An optional dictionary for the request body that gets url-encoded.
          files: An optional dictionary of files to upload as part of a multipart request.
          json: An optional value for the request body that gets json-encoded.
          auth: An optional function that can be used to add auth
            headers to the request.
        """
        if data is not None and json is not None or files is not None and json is not None:
            raise RuntimeError("either 'data'/'files' or 'json' should be provided, not both")

        request = Request(
            method=method.upper(),
            path=path,
            headers=headers,
            params=params,
        )
        if body is not None:
            request.headers["content-length"] = f"{len(body)}"
            request.body_file = BytesIO(body)

        elif files is not None:
            boundary = "--" + "".join(random.choice(string.ascii_letters) for _ in range(30))
            request.body_file = content = BytesIO()

            content.write(f"{boundary}\r\n".encode())
            for field_name, file_spec in files.items():
                file_type = "application/octet-stream"
                if len(file_spec) == 3:
                    file_name, file_type, file_stream = file_spec  # type: ignore
                else:
                    file_name, file_stream = file_spec  # type: ignore

                content.write(f"--{boundary}\r\n".encode())
                content.write(f'content-disposition: multipart/form-data; name="{field_name}"; filename="{file_name}"\r\n'.encode())
                content.write(f'content-type: {file_type}\r\n'.encode())
                content.write(b"\r\n")
                content.write(file_stream.read())
                content.write(b"\r\n")

            for field_name, field_value in (data or {}).items():
                content.write(f"--{boundary}\r\n".encode())
                content.write(f'content-disposition: multipart/form-data; name="{field_name}"\r\n'.encode())
                content.write(b"\r\n")
                content.write(field_value.encode())
                content.write(b"\r\n")

            content.write(f"--{boundary}--\r\n".encode())
            request.headers["content-type"] = f"multipart/form-data; boundary={boundary}"
            request.headers["content-length"] = str(content.tell())
            content.seek(0, os.SEEK_SET)

        elif data is not None:
            request_content = urlencode(data).encode("utf-8")
            request.headers["content-type"] = "application/x-www-form-urlencoded"
            request.headers["content-length"] = f"{len(request_content)}"
            request.body_file = BytesIO(request_content)

        elif json is not None:
            request_content = to_json(json).encode("utf-8")
            request.headers["content-type"] = "application/json; charset=utf-8"
            request.headers["content-length"] = f"{len(request_content)}"
            request.body_file = BytesIO(request_content)

        if auth is not None:
            request = auth(request)

        response = Response(HTTP_200)

        def start_response(status, response_headers, exc_info=None):  # type: ignore
            nonlocal response
            response.status = status
            response.headers = Headers(dict(response_headers))

        try:
            environ = to_environ(request)
            if prepare_environ:  # pragma: no cover
                environ = prepare_environ(environ)

            chunks = self.app(environ, start_response)
        except NoMoreData:
            chunks = []

        if response.headers.get("transfer-encoding") == "chunked":
            response.stream = cast(BinaryIO, chunks)
            return TestResponse(response)

        for chunk in chunks:
            response.stream.write(chunk)

        response.stream.seek(0)
        return TestResponse(response)

    def __getattr__(self, name: str) -> Any:
        if name in HTTP_METHODS:
            return partial(self.request, name)
        raise AttributeError(f"unknown attribute {name}")



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, BinaryIO, Callable, Dict, List, NewType, Tuple, Union, no_type_check

import typing_inspect

#: An alias representing a WSGI environment dictionary.
Environ = Dict[str, Any]

#: An alias representing a WSGI start_response callback.
StartResponse = Callable[[str, List[Tuple[str, str]], Any], None]

#: The type of middleware functions.
Middleware = Callable[[Callable[..., Any]], Callable[..., Any]]

#: The request method.
Method = NewType("Method", str)

#: The request URI scheme.
Scheme = NewType("Scheme", str)

#: The request hostname.
Host = NewType("Host", str)

#: The request port.
Port = NewType("Port", int)

#: The request query string.
QueryString = NewType("QueryString", str)

#: A query string parameter.
QueryParam = NewType("QueryParam", str)

#: A header.
Header = NewType("Header", str)

#: A file-like object representing the request input data.
RequestInput = NewType("RequestInput", BinaryIO)

#: A bytestring representing the request input data.
RequestBody = NewType("RequestBody", bytes)

#: Parsed request data.
RequestData = NewType("RequestData", Dict[str, Any])


@no_type_check
def flatten_type(t):
    # This is mainly a shim for Python 3.6 support.  It takes a tuple
    # like (typing.Dict, str, str) and generates typing.Dict[str,
    # str].  It's kind of awful and we should get rid of it as soon as
    # we can.
    if isinstance(t, tuple):
        return t[0][tuple(flatten_type(t) for t in t[1:])]

    return t


def get_args(annotation: Any) -> Any:
    # This is a safe version of get_args that works the same on Python
    # 3.6 and 3.7 by ensuring that expanded type arguments are merged
    # into their original type.
    arguments = list(typing_inspect.get_args(annotation))
    for i, argument in enumerate(arguments[:]):
        if isinstance(argument, tuple):
            arguments[i] = argument[0][tuple(flatten_type(t) for t in argument[1:])]

    return arguments


def extract_optional_annotation(annotation: Any) -> Tuple[bool, Any]:
    """Returns a tuple denoting whether or not the annotation is an
    Optional type and the inner annotation.
    """
    if typing_inspect.is_union_type(annotation):
        args = get_args(annotation)
        inner = [arg for arg in args if not isinstance(arg, type) or not issubclass(arg, type(None))]
        if inner != args:
            return True, Union[tuple(inner)]
    return False, annotation


def is_optional_annotation(annotation: Any) -> bool:
    """Returns True if the given annotation represents an Optional type.
    """
    is_optional, _ = extract_optional_annotation(annotation)
    return is_optional



# This file is a part of molten.
#
# Copyright (C) 2018 CLEARTYPE SRL <bogdan@cleartype.io>
#
# molten is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# molten is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from inspect import Parameter
from typing import Any, Dict, Optional

#: Canary value representing missing values.
Missing = object()


class Settings(Dict[str, Any]):
    """A dictionary of settings.
    """

    def deep_get(self, path: str, default: Optional[Any] = None) -> Optional[Any]:
        """Look up a deeply-nested setting by its path.

        Examples:

          >>> settings = Settings({"a": {"b": [{"c": 42}]}})
          >>> settings.deep_get("a.b.0.c")
          42

        Raises:
          TypeError: When attempting to index into a primitive value
            or when indexing a list with a string value rather than an
            integer.

        Parameters:
          path: A dot-separated string representing the path to the value.
          default: The value to return if the path cannot be traversed.
        """
        root = self
        names = path.split(".")
        for name in names:
            if isinstance(root, list):
                try:
                    root = root[int(name)]

                except (IndexError, ValueError):
                    raise TypeError(f"invalid index '{name}' for list {root!r}")

            elif isinstance(root, dict):
                root = root.get(name, Missing)

            else:
                raise TypeError(f"value {root!r} at subpath '{name}' is not a list or a dict")

            if root is Missing:
                return default

        return root

    def strict_get(self, path: str) -> Any:
        """Get a required setting.

        Raises:
          RuntimeError: If the value for that setting cannot be found.
        """
        value = self.deep_get(path, Missing)
        if value is Missing:
            raise RuntimeError(f"Cannot find required setting at path {path!r}.")

        return value


class SettingsComponent:
    """A component for settings that you build at app init time.

    Examples:

        >>> settings = Settings({"database_engine_dsn": "sqlite://"})
        >>> app = App(components=[SettingsComponent(settings)])

    """

    __slots__ = ["settings"]

    is_cacheable = True
    is_singleton = True

    def __init__(self, settings: Settings) -> None:
        self.settings = settings

    def can_handle_parameter(self, parameter: Parameter) -> bool:
        return parameter.annotation is Settings

    def resolve(self) -> Settings:
        return self.settings



import os

from setuptools import setup


def rel(*xs):
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), *xs)


with open(rel("README.md")) as f:
    long_description = f.read()


with open(rel("molten", "__init__.py"), "r") as f:
    version_marker = "__version__ = "
    for line in f:
        if line.startswith(version_marker):
            _, version = line.split(version_marker)
            version = version.strip().strip('"')
            break
    else:
        raise RuntimeError("Version marker not found.")


dependencies = [
    "typing-extensions>=3.6,<4.0",
    "typing-inspect>=0.3.1,<0.4",
]

extra_dependencies = {}
extra_dependencies["all"] = dependencies
extra_dependencies["dev"] = extra_dependencies["all"] + [
    # Docs
    "alabaster>0.7",
    "sphinx<1.8",
    "sphinxcontrib-napoleon",

    # Linting
    "flake8",
    "flake8-bugbear",
    "flake8-quotes",
    "isort",
    "mypy",

    # Misc
    "bumpversion>0.5,<0.6",

    # Contrib
    "dramatiq[rabbitmq]>1.3,<2.0",
    "gevent>1.3,<2.0",
    "gunicorn>19.8",
    "jinja2>=2.10,<3.0",
    "msgpack>0.5,<0.6",
    "prometheus-client>=0.2,<0.3",
    "sqlalchemy>1.2,<2.0",
    "toml>0.9,<0.10",
    "wsgicors>=0.7,<0.8",

    # Testing
    "pytest",
    "pytest-cov",
]

setup(
    name="molten",
    version=version,
    author="Bogdan Popa",
    author_email="bogdan@cleartype.io",
    description="A minimal, extensible, fast and productive API framework.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=[
        "molten",
        "molten.contrib",
        "molten.http",
        "molten.openapi",
        "molten.openapi.templates",
        "molten.testing",
        "molten.validation",
    ],
    include_package_data=True,
    install_requires=dependencies,
    python_requires=">=3.6",
    extras_require=extra_dependencies,
    classifiers=[
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)
